"""
   SDC Engine service model module
"""
from sqlalchemy import func
from sqlalchemy.orm import query_expression
from sqlalchemy.sql.functions import _FunctionGenerator
from sdc_helpers.models.service import Service as BaseService


class Service(BaseService):
    """
       SDC Engine Service model class
    """
    # pylint: disable=too-few-public-methods
    extend_existing = True

    maintenance_in_progress = query_expression()

    @classmethod
    def maintenance_in_progress_expression(cls) -> _FunctionGenerator:
        """
            SQL expression to get the maintenance_in_progress from properties

            returns:
                function (_FunctionGenerator): The JSON_EXTRACT SQL function

            Example SQL expression:

                JSON_EXTRACT(
                    properties,
                    "$.maintenance_in_progress"
                )
        """
        return func.json_extract(
            Service.properties,
            '$.maintenance_in_progress'
        )



"""
    Sagemaker Engine Helpers module used to manage
    and get information on client subscription
"""
import json
from sdc_engine_helpers.engine_query_helper import EngineQueryHelper
from sdc_engine_helpers.models.subscription import Subscription

class SagemakerEngineHelpers(EngineQueryHelper):
    """Extension on the EngineQueryHelper"""

    def get_solution(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            from_cache: bool = True,
            for_update: bool = False,
            **kwargs
    ):

        # get solution
        solution, index = super().get_solution(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            from_cache=from_cache,
            for_update=for_update,
            **kwargs
        )

        if not solution:
            raise IndexError(
                (
                    'ClientError: No Solution matching {} '
                    'found for Engine = "{}" '
                    'And client = "{}" '.format(json.dumps(kwargs), engine_slug, client_id)
                )
            )

        return solution, index

    def get_dataset(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            from_cache: bool = True,
            for_update: bool = False,
            **kwargs
    ):
        dataset, index = super().get_dataset(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            from_cache=from_cache,
            for_update=for_update,
            **kwargs
        )

        if not dataset:
            raise IndexError(
                (
                    "ClientError: No Dataset matching {} "
                    "found in Engine = '{}' "
                    "for client = '{}'".format(json.dumps(kwargs), engine_slug, client_id)
                )
            )

        return dataset, index

    def get_item_from_lookup(
            self,
            *,
            client_id: int,
            service_id: int,
            engine: str,
            from_cache: bool = True,
            for_update: bool = False,
            label: str,
            key: str
    ):
        """
            Get a specific item from lookup

            args:
                client_id (int): The client id of the subscription
                service_id (int): The service id of the subscription
                engine (str): Subscription property engine
                from_cache (bool): Retrieve the datasets from cache - Default True
                label (str): Label describing what the dataset is for
                key (str): Dataset key to lookup

            returns:
                result (dict): The result of the lookup
        """
        dataset_kwargs = {'label': label}

        dataset, _ = self.get_dataset(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine,
            from_cache=from_cache,
            for_update=for_update,
            **dataset_kwargs
        )

        return dataset.get("data", {}).get(str(key))

    def get_campaign(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            from_cache: bool = True,
            for_update: bool = False,
            **kwargs
    ):
        campaign, index = super().get_campaign(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            from_cache=from_cache,
            for_update=for_update,
            **kwargs
        )

        if not campaign:
            raise IndexError(
                (
                    "ClientError: No Campaign matching {} "
                    "found for Engine = '{}' "
                    "And client = '{}' ".format(json.dumps(kwargs), engine_slug, client_id)
                )
            )

        return campaign, index

    @staticmethod
    def update_engine_entity_on_index(
            *,
            subscription: Subscription,
            engine_index: int,
            entity_key: str,
            entity_index: int,
            value
    ):
        """
            Update a specific entity in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                entity_key (str): The key of the entity in the engine
                value: The data to set the entity key node to
        """
        subscription.properties['engines'][engine_index][entity_key][entity_index] = value

    def update_dataset(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            config: dict,
            from_cache: bool = True,
            **kwargs
    ):
        """
            Update specific dataset in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                value: The data to set the entity key node to
        """

        # get subscription for engine
        subscription = self.get_subscription(
            client_id=client_id,
            service_id=service_id,
            from_cache=False,
            for_update=True
        )

        # get dataset index
        dataset, dataset_index = self.get_dataset(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            from_cache=from_cache,
            for_update=False,
            **kwargs
        )
        # update config
        dataset.update(config)

        # get engine_index
        _, engine_index = self.get_engine_entity(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            entity_key='datasets',
            from_cache=False,
            for_update=False
        )

        # update entity
        self.update_engine_entity_on_index(
            subscription=subscription,
            engine_index=engine_index,
            entity_key='datasets',
            entity_index=dataset_index,
            value=dataset
        )

    def update_campaign(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            config: dict,
            from_cache: bool = True,
            **kwargs
    ):
        """
            Update specific campaign in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                value: The data to set the entity key node to
        """
        # get subscription for engine
        subscription = self.get_subscription(
            client_id=client_id,
            service_id=service_id,
            from_cache=False,
            for_update=True
        )

        # get campaign index
        campaign, campaign_index = self.get_campaign(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            from_cache=from_cache,
            for_update=False,
            **kwargs
        )

        campaign.update(config)

        # get engine_index
        _, engine_index = self.get_engine_entity(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            entity_key='campaigns',
            from_cache=False,
            for_update=False
        )
        # update the engine campaign
        self.update_engine_entity_on_index(
            subscription=subscription,
            engine_index=engine_index,
            entity_key='campaigns',
            entity_index=campaign_index,
            value=campaign
        )

    def update_solution(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            config: dict,
            from_cache: bool = True,
            **kwargs
    ):
        """
            Update specific solution in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                value (dict): dataset configs to set the entity key node to
        """
        # get subscription for engine
        subscription = self.get_subscription(
            client_id=client_id,
            service_id=service_id,
            from_cache=False,
            for_update=True
        )

        solution, solution_index = self.get_solution(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            from_cache=from_cache,
            for_update=False,
            **kwargs
        )
        solution.update(config)
        # get engine_index
        _, engine_index = self.get_engine_entity(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            entity_key='solutions',
            from_cache=False,
            for_update=False
        )
        # update the engine solution
        self.update_engine_entity_on_index(
            subscription=subscription,
            engine_index=engine_index,
            entity_key='solutions',
            entity_index=solution_index,
            value=solution
        )



"""
    Personalize event manager module
"""
import json
import time
import boto3
from sdc_engine_helpers.personalize.manager import Manager


class EventManager(Manager):
    """
        Manage real-time events to AWS personalize for a given set of parameters
    """
    # pylint: disable=raise-missing-from
    def track_event(self):
        """
            Track the personalise event from the event
            dictionary provided by lambda

            returns:
                results (dict): A dict for logging the event

        """
        client = self.get_client()

        event_type = self.parameters.get('eventType', 'listingView')

        dataset_kwargs = {'label': 'tracking'}

        dataset, _ = self.query_helper.get_dataset(
            client_id=client.id,
            service_id=self.service.id,
            engine_slug=self.engine_slug,
            **dataset_kwargs
        )

        if dataset is None:
            raise Exception('ServerError: Could not determine tracking dataset for this client')

        tracking_id = dataset.get('identifier', None)
        if tracking_id is None:
            raise Exception('ServerError: Identifier is not populated in tracking dataset')

        item_id = self.parameters.get('itemId')
        user_id = self.parameters.get('userId')
        timestamp = self.parameters.get('timestamp', int(time.time()))

        try:
            personalize_events = boto3.client('personalize-events')

            personalize_events.put_events(
                trackingId=tracking_id,
                userId=user_id,
                sessionId=user_id,
                eventList=[
                    {
                        'sentAt': timestamp,
                        'eventType': event_type,
                        'properties': json.dumps(
                            {
                                'itemId': item_id,
                            }
                        )
                    }
                ]
            )
        except Exception as ex:
            raise Exception(
                'ServerError: {exception}'.format(exception=str(ex))
            )

        print('Successful put event')

        results = {
            'client': client.name,
            'context': self.parameters.get('context', None),
            'user_id': user_id,
            'item_id': item_id
        }
        print(results)

        return results



"""
    Personalize abstract manager module
"""
from sdc_helpers.models.client import Client
from sdc_helpers.redis_helper import RedisHelper
from sdc_engine_helpers.engine_query_helper import EngineQueryHelper

class Manager:
    """
        Abstract manager class to be inherited by various Personalize managers
    """
    service = None
    service_for_update = False
    engine_slug = None
    query_helper = None
    redis_helper = None
    parameters = None
    # flag to determine if maintenance was performed
    can_perform_maintenance = False

    def __init__(self, **kwargs):
        query_helper_kwargs = dict(filter(lambda item: item[0] == 'rds_config', kwargs.items()))
        redis_helper_kwargs = dict(filter(lambda item: item[0] == 'redis_config', kwargs.items()))
        self.query_helper = EngineQueryHelper(**query_helper_kwargs)
        self.redis_helper = RedisHelper(**redis_helper_kwargs)
        self.service_slug = kwargs.pop('service_slug', 'recommend')
        self.service = self.query_helper.get_service(
            slug=self.service_slug,
            for_update=self.service_for_update
        )
        self.engine_slug = kwargs.pop('engine_slug', 'personalize')
        self.parameters = kwargs

    def __del__(self):
        if self.can_perform_maintenance:
            # reset maintenance_in_progress
            self.service = self.query_helper.get_service(
                slug=self.service_slug,
                for_update=True
            )

            self.service.properties['maintenance_in_progress'] = None
            # update service
            self.query_helper.update(
                model=self.service,
                audit=False
            )
        del self.query_helper
        del self.redis_helper

    def check_maintenance_is_allowed(self) -> bool:
        """
            checks if maintenance is currently in progress for a given service.
            if not, returns True allowing the current manager to take ownership and run
            maintenance else returns False to exit gracefully.

            Returns:
                bool: Whether current manager can perform maintenance.
        """
        allow_maintenance = False

        if (
                isinstance(self.service.properties, dict) and
                self.service.properties.get('maintenance_in_progress', None) is None
        ):
            allow_maintenance = True
        elif (
                self.service.properties is None
        ):
            # setting empty properties field as a dict in python model
            self.service.properties = {}
            allow_maintenance = True

        if allow_maintenance:
            # no maintenance in progress
            # run maintenance
            self.service.properties['maintenance_in_progress'] = self.engine_slug

            self.query_helper.update(
                model=self.service,
                audit=False
            )

        return allow_maintenance

    def get_client(self) -> Client:
        """
            Determine the client with the supplied parameters

            returns:
                client (Client): The determined client

        """
        client = None

        client_id = self.parameters.get('client_id', None)
        if client_id is not None:
            client = self.query_helper.get_client(client_id=client_id)
        else:
            api_key_id = self.parameters.get('api_key_id', None)
            if api_key_id is not None:
                client = self.query_helper.get_client(api_key_id=api_key_id)

        if client is None:
            raise Exception('ClientError: Could not determine client for this request')

        return client



"""
    Maintenance manager module
        - performs maintenance on solution, campaign, datasets
        - calls solution, campaign, datasets which interact with AWS
        - gets current states in database
        - updates states in database

    Issues:
     - boto3 does not allow subclassing & builds internal classes at runtime:
        https://boto3.amazonaws.com/v1/documentation/api/latest/guide/events.html
"""
import boto3
from sqlalchemy.exc import InternalError
from botocore.exceptions import ClientError as botocore_clienterror
from sdc_helpers.slack_helper import SlackHelper
from sdc_helpers import decorators as global_decorators
from sdc_engine_helpers.sagemaker.utils import create_resource_name
from sdc_engine_helpers.models.subscription import Subscription
from sdc_engine_helpers.date_utils import DateUtils
from sdc_engine_helpers import decorators as engine_decorators
from sdc_engine_helpers.sagemaker.sagemaker_query_helpers import \
    SagemakerEngineHelpers
from sdc_engine_helpers.sagemaker.maintenance.campaign_state_manager import \
    CampaignStateManager
from sdc_engine_helpers.sagemaker.maintenance.solution_state_manager import \
    SolutionStateManager
from sdc_engine_helpers.sagemaker.maintenance.dataset_state_manager import \
    DatasetStateManager

class MaintenanceManager:
    """
        Manage AWS SDC Engine solutions and campaigns
    """
    # pylint: disable=too-many-instance-attributes, raise-missing-from
    # flag to determine if maintenance was performed
    current_client = None
    can_perform_maintenance = False
    # get helpers
    date_utils = DateUtils()
    slack_helper = SlackHelper()

    # set up states
    solution_job_status = {
        'active': ['Completed']
    }

    def __init__(self, **kwargs):
        # get service slug
        self.service_slug = kwargs.get('service_slug', 'recommend')
        # get rds config
        rds_config = kwargs.get('rds_config', {})
        if not isinstance(rds_config, dict):
            raise TypeError('Only dict is supported for rds_config')

        # get redis config
        redis_config = kwargs.get('redis_config', {})
        if not isinstance(redis_config, dict):
            raise TypeError('Only dict is supported for redis_config')

        self.engine_slug = kwargs.get('engine_slug', 'alice')
        # instantiate query helpers
        self.query_helper = SagemakerEngineHelpers(
            rds_config=rds_config,
            redis_config=redis_config
        )
        # get service config for service slug
        self.service = self.query_helper.get_service(
            slug=self.service_slug,
            for_update=True
        )
        # create instance of sagemaker
        self.sagemaker = boto3.client('sagemaker')
        # init state managers
        self.solution_state_manager = SolutionStateManager()
        self.campaign_state_manager = CampaignStateManager()
        self.dataset_state_manager = DatasetStateManager()
        process = self.query_helper.get_process(slug='sagemaker-maintenance')
        if process is not None:
            self.query_helper.process_id = process.id

    def __del__(self):
        # finally close query helper
        if self.can_perform_maintenance:
            # reset maintenance_in_progress
            self.service = self.query_helper.get_service(
                slug=self.service_slug,
                for_update=True
            )

            self.service.properties['maintenance_in_progress'] = None
            # update service
            self.query_helper.update(
                model=self.service,
                audit=False
            )
        del self.query_helper

    def check_maintenance_is_allowed(self) -> bool:
        """
            checks if maintenance is currently in progress for a given service.
            if not, returns True allowing the current manager to take ownership and run
            maintenance else returns False to exit gracefully.

            Returns:
                bool: Whether current manager can perform maintenance.
        """
        allow_maintenance = False

        if (
                isinstance(self.service.properties, dict) and
                self.service.properties.get('maintenance_in_progress', None) is None
        ):
            allow_maintenance = True
        elif (
                self.service.properties is None
        ):
            # setting empty properties field as a dict in python model
            self.service.properties = {}
            allow_maintenance = True

        if allow_maintenance:
            # no maintenance in progress
            # run maintenance
            self.service.properties['maintenance_in_progress'] = self.engine_slug

            self.query_helper.update(
                model=self.service,
                audit=False
            )

        return allow_maintenance

    def perform_maintenance(self):
        """
            Execute maintenance that performs a set of tasks for engines
            These are specifically Update, refresh or do nothing.

            Return:
                (bool) : the outcome of maintenance "updated", "not_updated", "blocked"
        """
        maintenance_outcome = "not_updated"

        # check if maintenance can be performed
        self.can_perform_maintenance = self.check_maintenance_is_allowed()

        if self.can_perform_maintenance:
            # start maintenance
            for client in self.query_helper.get_clients():
                maintenance_mode, _ = self.query_helper.get_engine_entity(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=self.engine_slug,
                    entity_key='maintenance_mode',
                    from_cache=False
                )

                if maintenance_mode is None or maintenance_mode == 'off':
                    continue

                # get subscription for engine
                subscription = self.query_helper.get_subscription(
                    client_id=client.id,
                    service_id=self.service.id,
                    from_cache=False,
                    for_update=True
                )

                self.current_client = client
                solution_updated = self.perform_solution_maintenance(
                    client=client,
                    engine_slug=self.engine_slug,
                    maintenance_mode=maintenance_mode
                )

                campaign_updated = self.perform_campaign_maintenance(
                    client=client,
                    engine_slug=self.engine_slug,
                    maintenance_mode=maintenance_mode
                )

                datasets_updated = self.perform_dataset_maintenance(
                    client=client,
                    engine_slug=self.engine_slug,
                    maintenance_mode=maintenance_mode
                )

                if any([solution_updated, campaign_updated, datasets_updated]):
                    # when any of the above have been updated
                    maintenance_outcome = self.perform_session_update(
                        subscription=subscription
                    )
        else:
            maintenance_outcome = "blocked"

        return maintenance_outcome

    @engine_decorators.retry_handler(
        exceptions=(Exception, InternalError),
        total_tries=4,
        initial_wait=0.5,
        backoff_factor=2
    )
    def perform_session_update(self, subscription: Subscription) -> bool:
        """
            Perform the session update in the database to commit update to database

            Returns:
                updated (bool): whether the database session was updated
        """
        print("Updating subscription")
        self.query_helper.update(
            model=subscription
        )
        print("clearing cache")
        self.query_helper.flush_subscription_properties_cache(
            subscription=subscription
        )

        return "updated"


    @global_decorators.general_exception_handler(
        logger=slack_helper,
        exceptions=(botocore_clienterror, Exception),
        should_raise=False
    )
    def perform_dataset_maintenance(
            self, *,
            client: object,
            engine_slug: str,
            maintenance_mode: str
    ) -> dict:
        """
            Performs dataset update maintainence

            - get state
            - for each dataset, check state for vendor
        """
        datasets_updated = False
        curr_state = {}
        if maintenance_mode != 'off':
            # get datasets
            datasets, _ = self.query_helper.get_engine_entity(
                client_id=client.id,
                service_id=self.service.id,
                engine_slug=engine_slug,
                entity_key='datasets',
                from_cache=False,
                for_update=True
            )

            if datasets is None:
                return datasets_updated

            for _, dataset in enumerate(datasets):

                label = dataset.get('label', None)
                if label is None:
                    # skip this dataset
                    continue

                new_dataset = {}

                # get solution from db
                campaign, _ = self.query_helper.get_campaign(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=engine_slug,
                    from_cache=False,
                    for_update=True,
                    recipe=dataset['recipe']
                )
                # get dataset
                curr_state = self.dataset_state_manager.get_state(
                    dataset=dataset,
                    campaign=campaign
                )

                if curr_state.get('should_update') is True:

                    print(
                        'Updating dataset version for client: {client_name}'
                        .format(client_name=client.name)
                    )

                    # build cache key
                    cache_key = create_resource_name(
                        context="recommendation",
                        function=dataset['label'],
                        additional_details=[
                            self.service.id, client.id, engine_slug
                        ]
                    )

                    new_dataset = self.dataset_state_manager.update_state(
                        dataset=dataset,
                        campaign=campaign,
                        cache_key=cache_key
                    )

                    # update dataset
                    self.query_helper.update_dataset(
                        client_id=client.id,
                        service_id=self.service.id,
                        engine_slug=engine_slug,
                        label=dataset['label'],
                        config=new_dataset
                    )

                    dataset_label = dataset['label']
                    print(
                        'Successfully updated dataset version: {dataset_label}'
                        .format(dataset_label=dataset_label)
                    )

                    datasets_updated = True

        return datasets_updated


    @global_decorators.general_exception_handler(
        logger=slack_helper,
        exceptions=(botocore_clienterror, Exception),
        should_raise=False
    )
    def perform_solution_maintenance(
            self,
            *,
            client: dict,
            engine_slug: str,
            maintenance_mode: str
    ) -> dict:
        """
            Perform the required maintenance on a list of solutions

            args:
                client (dict): The client this maintenance is being performed for
                engine_slug (str): Slug of the engine involved
                maintenance_mode (str): The maintenance mode of the subscription
                                        (full/database_only)

            returns:
                result (dict): A dictionary with an updated flag and resultant solutions

        """
        solutions_updated = False
        curr_state = {}
        # get solutions
        solutions, _ = self.query_helper.get_engine_entity(
            client_id=client.id,
            service_id=self.service.id,
            engine_slug=engine_slug,
            entity_key='solutions',
            from_cache=False,
            for_update=True
        )

        if solutions is None:
            return solutions_updated

        for _, solution in enumerate(solutions):

            # The solution arn is required to continue maintenance for this solution
            arn = solution.get('arn', None)

            if arn is None:
                continue

            # get current state
            curr_state = self.solution_state_manager.get_state(
                solution=solution
            )

            if maintenance_mode == 'full' and curr_state.get('should_create_version') is True:
                print(
                    'Building solution version for client: {client_name}'
                    .format(client_name=client.name)
                )

                # train a new model if should_create_version True
                new_solution = self.solution_state_manager.update_state(
                    solution=solution,
                    client_code=client.code,
                    service_slug=self.service.slug,
                    engine_slug=engine_slug,
                )

                # get solution arn
                solution_arn = new_solution.get('arn')
                print(
                    'Successfully started creation of solution version: {solution_version_arn}'
                    .format(solution_version_arn=solution_arn)
                )

                # update immedidately and change status
                self.query_helper.update_solution(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=engine_slug,
                    recipe=solution['recipe'],
                    config=new_solution
                )

                solutions_updated = True

            elif curr_state.get('should_refresh_version_status') is True:
                # update state step >> solution manager
                solution_arn = curr_state.get('arn')
                new_version_status = curr_state.get('new_version_status')

                print(
                    'Refreshing version status on solution: {arn} for client: {client_name}'
                    ' to {status}'
                    .format(
                        arn=arn,
                        client_name=client.name,
                        status=new_version_status
                    )
                )
                # update immedidately and change status
                self.query_helper.update_solution(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=engine_slug,
                    recipe=solution['recipe'],
                    config={
                        'version_status': new_version_status
                    }
                )

                solutions_updated = True

        return solutions_updated


    @global_decorators.general_exception_handler(
        logger=slack_helper,
        exceptions=(botocore_clienterror, Exception),
        should_raise=False
    )
    def perform_campaign_maintenance(
            self, *,
            client: dict,
            engine_slug: str,
            maintenance_mode: str
    ) -> dict:
        """
            Perform the required maintenance on a list of campaigns

            args:
                client (dict): The client this maintenance is being performed for
                engine_slug (str): Slug of the engine involved
                maintenance_mode (str): The maintenance mode of the subscription
                                        (full/database_only)

            returns:
                result (dict): Dictionary containing:
                               1) campaigns_updated
                               2) Resultant campaigns list

        """
        campaigns_updated = False
        curr_state = {}
        # get campaigns for this engine
        campaigns, _ = self.query_helper.get_engine_entity(
            client_id=client.id,
            service_id=self.service.id,
            engine_slug=engine_slug,
            entity_key='campaigns',
            from_cache=False,
            for_update=True
        )

        if campaigns is None:
            return campaigns_updated

        for _, campaign in enumerate(campaigns):
            # The campaign arn is required to continue maintenance for this campaign
            arn = campaign.get('arn', None)

            if arn is None:
                continue

            # get related-listings solution
            solution, _ = self.query_helper.get_solution(
                client_id=client.id,
                service_id=self.service.id,
                engine_slug=engine_slug,
                from_cache=False,
                for_update=True,
                recipe=campaign['recipe']
            )

            curr_state = self.campaign_state_manager.get_state(
                campaign=campaign,
                solution=solution
            )

            if maintenance_mode == 'full' and curr_state.get('should_update') is True:

                # update campaign state >> campaign_state_manager
                latest_solution_version_arn = curr_state.get(
                    'latest_solution_version_arn'
                )

                print(
                    'Updating campaign for client: {client_name} to solution version: {arn}'
                    .format(
                        client_name=client.name,
                        arn=latest_solution_version_arn
                    )
                )

                new_campaign = self.campaign_state_manager.update_state(
                    campaign=campaign,
                    solution=solution,
                    client_code=client.code,
                    service_slug=self.service.slug,
                    engine_slug=engine_slug
                )

                print('Successfully started updating campaign: {arn}'.format(arn=arn))

                # try and update else fail
                self.query_helper.update_campaign(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=engine_slug,
                    recipe=campaign['recipe'],
                    config=new_campaign
                )

                campaigns_updated = True

            elif curr_state.get('should_refresh_status') is True:
                # get current version arn
                latest_solution_version_arn = curr_state.get(
                    'latest_solution_version_arn'
                )
                new_status = curr_state.get('new_status')
                last_refreshed_at = curr_state.get('last_refreshed_at')

                print(
                    'Refreshing status on campaign: {arn} for client: {client_name} to {status}'
                    .format(
                        arn=arn,
                        client_name=client.name,
                        status=new_status
                    )
                )
                new_status = curr_state.get('new_status')
                last_refreshed_at = curr_state.get('last_refreshed_at')

                # solution refresh to account for endpoint roll-back
                # when sagemaker endpoint update fails, it rolls back
                # as such campaign in db and vendor will differ
                # this is a roll back
                refresher_solution = {
                    'status':new_status,
                    'solution_version_arn':latest_solution_version_arn,
                    'last_updated_at': last_refreshed_at
                }
                # update status in db
                self.query_helper.update_campaign(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=engine_slug,
                    recipe=campaign['recipe'],
                    config=refresher_solution
                )

                campaigns_updated = True

        return campaigns_updated



"""
    Sagemaker Utils
        - this is a fork of sagemaker.utils found here:
        https://github.com/aws/sagemaker-python-sdk/blob/master/src/sagemaker/utils.py
"""
import time
from urllib.parse import urlparse

def sagemaker_timestamp():
    """Return a timestamp with millisecond precision."""
    moment = time.time()
    moment_ms = repr(moment).split(".")[1][:3]
    return time.strftime("%Y-%m-%d-%H-%M-%S-{}".format(moment_ms), time.gmtime(moment))

def sagemaker_short_timestamp():
    """Return a timestamp that is relatively short in length"""
    return time.strftime("%y%m%d-%H%M")

def name_from_base(base, max_length=63, short=False):
    """Append a timestamp to the provided string.
    This function assures that the total length of the resulting string is
    not longer than the specified max length, trimming the input parameter if
    necessary.
    Args:
        base (str): String used as prefix to generate the unique name.
        max_length (int): Maximum length for the resulting string.
        short (bool): Whether or not to use a truncated timestamp.
    Returns:
        str: Input parameter with appended timestamp.
    """
    timestamp = sagemaker_short_timestamp() if short else sagemaker_timestamp()
    trimmed_base = base[: max_length - len(timestamp) - 1]
    return "{}-{}".format(trimmed_base, timestamp)

def parse_s3_url(url: str):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3
    scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            str: S3 bucket name str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != "s3":
        raise ValueError("Expecting 's3' scheme, got: {} in {}.".format(parsed_url.scheme, url))
    return parsed_url.netloc, parsed_url.path.lstrip("/")

def create_resource_name(
        *,
        context: str,
        function: str,
        additional_details: list = None
):
    """
        Constructs a structured string name, separated by a "-" delimiter.

        args:
            service (str):
            group (str): group name which speaks to function of resource
            prefix (list): (optiional) list of args containing prefixes to add in order

        Return:
            (str) : name of the form {service}-{prefixes}-{group}

        example:
            >> cache_key = SagemakerEngineHelpers.build_resource_name(
                service='recommend', group='lookup', prefix=['prefix1', 'prefix2', 'prefix3'])
            >> print(cache_key)
            'recommend-prefix1-prefix2-prefix3-lookup'
    """
    # join prefixes
    if additional_details is not None:
        additional_details = [str(p) for p in additional_details]
        prefixes = "-".join(additional_details)

        # build cache key
        cache_key = '{context}-{function}-{prefixes}'.format(
            context=context,
            prefixes=prefixes,
            function=function
        )
    else:
        # build cache key
        cache_key = '{context}-{function}'.format(
            context=context,
            function=function
        )

    return cache_key.replace(" ", "-").replace("_", "-").lower()

def build_resource_tags(current_tags: list, new_tags: list) -> list:
    """
        given a list of tags, in the AWS sagemaker tag format
        looks for resource_tags and creates a new set of resource tags
        by adding tags given. Finally returns the new tags.

    Args:
        tags (list): list of resource tags to add

    Returns:
        list: new resource tags
    """
    new_resource_tags = []
    if current_tags not in [None, []]:
        new_resource_tags = current_tags
    for tag in new_tags:
        tag_key = tag.get('Key', None)
        if tag_key is None:
            raise Exception((
                "KeyError: cannot be None. "\
                "Make Sure you use \"Key\" & \"Value\" style AWS tags"
            ))

        for index, resource_tag in enumerate(current_tags):
            if tag_key.lower() == resource_tag.get('Key').lower():
                # remove if already exists
                new_resource_tags.pop(index)
                break
        # add new key to tags
        new_resource_tags.append(tag)

    return new_resource_tags



"""
    Campaign state manager module
"""
from datetime import datetime
import boto3
from botocore.exceptions import ClientError as botocore_clienterror
import sdc_helpers.utils as utils
from sdc_engine_helpers.sagemaker.utils import \
    build_resource_tags
from sdc_engine_helpers.date_utils import DateUtils
from sdc_engine_helpers.sagemaker.utils import name_from_base

class CampaignStateManager:
    """
        Given a database campaign object,

        1. get_state
            - call vendor API's to check if campaign needs to be updated
            OR
            - check if the campaign status in the database needs to be refreshed.
        2. update_state
            - call vendor API's to execute update workflow for campaign
            - update status of campaign in database

    """
    # pylint: disable=raise-missing-from
    # default tag to dev environment
    resource_tags = [
        {
            'Key': 'environment',
            'Value': 'dev'
        }
    ]
    date_utils = DateUtils()
    # vendor
    vendor_solution_arn_path = 'TrainingJobName'
    vendor_campaign_arn_path = 'EndpointName'
    vendor_campaign_status_path = 'EndpointStatus'
    vendor_solution_status_path = 'TrainingJobStatus'
    # input data key in TrainingJob
    vendor_campaign_solution_arn_path = 'ProductionVariants.0.VariantName'

    campaign_statuses = {
        'running':['Creating', 'Updating', 'SystemUpdating', 'Deleting'],
        'updating': ['Updating'],
        'failed':['Failed', 'RollingBack'],
        'danger':['Failed'],
        'blocking':[
            'Creating',
            'Updating',
            'SystemUpdating',
            'Deleting',
            'Failed',
            'RollingBack'
        ],
        'stopped':['OutOfService'],
        'active':['InService', None, 'None']
    }

    solution_statuses = {
        'blocking': ['InProgress', 'Stopping', 'Failed', 'Stopped'],
        'success': ['Completed', None, 'None']
    }

    dataset_statuses = {
        'active': ['Active', None, 'None'],
        'blocking': ['Updating']
    }

    def __init__(self):
        self.sagemaker = boto3.client('sagemaker')

    def sagemaker_create_model(
            self,
            training_job_name: str,
            container_image: str = None,
            resource_tags: list = None
    ) -> str:
        """
            Create model from a trained model job using Sagemaker API.
            if model exists, creation is skipped.

            note:
                - Model name is the training job name

            Args:
                TrainingJobName (str): Naming of sagemaker training job

            Returns:
                str: model name
        """
        model_name = training_job_name

        training_job_config = self.sagemaker.describe_training_job(
            TrainingJobName=training_job_name
        )

        primary_container = {
            'Image': training_job_config['AlgorithmSpecification']['TrainingImage'],
            'Mode': 'SingleModel',
            'ModelDataUrl': training_job_config['ModelArtifacts']['S3ModelArtifacts']
        }

        if container_image is not None:
            # override Image with provided image
            print((
                "Updating serving image to provided "
                "container image, {}".format(container_image)
            ))
            primary_container.update(
                {
                    'Image': container_image
                }
            )

        try:
            _ = self.sagemaker.create_model(
                ModelName=model_name,
                ExecutionRoleArn=training_job_config['RoleArn'],
                PrimaryContainer=primary_container,
                Tags=resource_tags
            )

        except botocore_clienterror as exception:
            if exception.response['Error']['Code'] == 'ValidationException':
                # handle no
                if 'Cannot create already existing model' in exception.response['Error']['Message']:
                    print((
                        "This Model already exists, "
                        "with name = {}. Using the already created model."
                        .format(model_name)
                    ))
                else:
                    # re-throw error
                    raise botocore_clienterror(
                        error_response=exception.operation_name,
                        operation_name=exception.response
                    )

        return model_name

    def sagemaker_update_campaign(
            self,
            endpoint_name: str,
            training_job_name: str,
            container_image: str = None,
            **kwargs
    ):
        """
            (Call vendor API) Creates a endpoint config
            and updates endpoint with this config

            Args:
                EndpointName (str): name of sagemaker endpoint
                TrainingJobName (str): name of sagemaker training job

            Returns:
                dict: contains names of new created resources
        """
        client_code = kwargs.get("client_code", None)
        service_slug = kwargs.get("service_slug", None)
        engine_slug = kwargs.get("engine_slug", None)
        # each endpoint configuration needs unique name
        endpoint_config_name = name_from_base(
            endpoint_name,
            max_length=63,
            short=False
        )
        resource_tags = build_resource_tags(
            current_tags=self.resource_tags,
            new_tags=[
                {
                    'Key': 'client',
                    'Value': client_code
                },
                {
                    'Key': 'environment',
                    'Value': 'prod'
                },
                {
                    'Key': 'engine',
                    'Value': engine_slug
                },
                {
                    'Key': 'service',
                    'Value': service_slug
                }
            ]
        )
        endpoint = self.sagemaker.describe_endpoint(
            EndpointName=endpoint_name
        )

        model_name = self.sagemaker_create_model(
            training_job_name=training_job_name,
            container_image=container_image,
            resource_tags=resource_tags
        )

        previous_endpoint_config = self.sagemaker.describe_endpoint_config(
            EndpointConfigName=endpoint['EndpointConfigName']
        )

        production_varients = previous_endpoint_config['ProductionVariants']
        # update the model name in production varients
        production_varients[0]['ModelName'] = model_name
        production_varients[0]['VariantName'] = model_name

        _ = self.sagemaker.create_endpoint_config(
            EndpointConfigName=endpoint_config_name,
            ProductionVariants=production_varients
        )

        _ = self.sagemaker.update_endpoint(
            EndpointName=endpoint_name,
            EndpointConfigName=endpoint_config_name
        )

        return {
            'EndpointName': endpoint_name,
            'TrainingJobName': training_job_name,
            'ModelName': model_name
        }

    def get_vendor_solution_version_arn(self, vendor_solution: dict):
        """
            Get vendor solution version status

            args:
                vendor_solution (dict): Solution from vendor

            returns:
                status (str): vendor solution version status

        """

        return utils.dict_query(
            dictionary=vendor_solution,
            path=self.vendor_solution_arn_path
        )

    def get_vendor_campaign_solution_version_arn(self, vendor_campaign: dict):
        """
            Get vendor campaign version status

            args:
                vendor_campaign (dict): Campaign from vendor

            returns:
                status (str): vendor campaign version status

        """

        return utils.dict_query(
            dictionary=vendor_campaign,
            path=self.vendor_campaign_solution_arn_path
        )

    def get_vendor_solution_status(self, vendor_solution: dict):
        """
            Get vendor solution version status

            args:
                vendor_solution (dict): Solution from vendor

            returns:
                status (str): vendor solution version status

        """

        return utils.dict_query(
            dictionary=vendor_solution,
            path=self.vendor_solution_status_path
        )

    def get_vendor_campaign_status(self, vendor_campaign: dict):
        """
            Get vendor campaign version status

            args:
                vendor_campaign (dict): Campaign from vendor

            returns:
                status (str): vendor campaign version status

        """

        return utils.dict_query(
            dictionary=vendor_campaign,
            path=self.vendor_campaign_status_path
        )

    def get_state(self, *, campaign: dict, solution: dict) -> dict:
        """
            Get the current state to action an update, refresh or do nothing
            set of actions.

            Args:
                campaign (dict): database campaign config
                solution (dict): database solution config
            Return:
                (dict): current state config

        """
        state = {
            'should_update': False,
            'should_refresh_status': False
        }

        # get vendor campaign
        vendor_campaign = self.sagemaker.describe_endpoint(
            EndpointName=campaign.get('arn')
        )
        # get vendor solution
        vendor_solution = self.sagemaker.describe_training_job(
            TrainingJobName=solution.get('arn')
        )

        vendor_solution_arn = self.get_vendor_solution_version_arn(
            vendor_solution=vendor_solution
        )

        state['should_update'] = self.should_update(
            campaign=campaign,
            vendor_campaign=vendor_campaign,
            vendor_solution=vendor_solution
        )

        if state['should_update']:
            state.update({
                'latest_solution_version_arn': vendor_solution_arn
            })

        else:
            # Check if the database arn needs to be updated
            arn = campaign.get('solution_version_arn', None)

            vendor_campaign_solution_version_arn = self.get_vendor_campaign_solution_version_arn(
                vendor_campaign=vendor_campaign
            )
            # Check if the database status needs to be updated
            status = campaign.get('status', None)

            vendor_campaign_status = self.get_vendor_campaign_status(
                vendor_campaign=vendor_campaign
            )

            if (
                    (vendor_campaign_status != status) &
                    (vendor_campaign_status not in self.campaign_statuses['blocking'])
            ):
                # update status, when it differs from vendor during updating/blocking status
                # get current refresh time, for keeping track
                current_time_at_update = datetime.strftime(
                    datetime.now(),
                    self.date_utils.get_mysql_date_format()
                )
                state['should_refresh_status'] = True
                state['new_status'] = vendor_campaign_status
                state['latest_solution_version_arn'] = vendor_campaign_solution_version_arn
                state['last_refreshed_at'] = current_time_at_update
            elif (
                    (vendor_campaign_solution_version_arn != arn) &
                    (vendor_campaign_status not in self.campaign_statuses['blocking'])
            ):
                # when not in updating/blocking status, and db solution arn is mismatched
                # with vendor solution arn
                # get current refresh time, for keeping track
                current_time_at_update = datetime.strftime(
                    datetime.now(),
                    self.date_utils.get_mysql_date_format()
                )
                state['should_refresh_status'] = True
                state['new_status'] = vendor_campaign_status
                state['latest_solution_version_arn'] = vendor_campaign_solution_version_arn
                state['last_refreshed_at'] = current_time_at_update

        return state

    def should_update(
            self,
            *,
            campaign: dict,
            vendor_campaign: dict,
            vendor_solution: dict
    ) -> bool:
        """
            Check whether a campaign should be updated when all of the following are true

            args:
                vendor_campaign(dict): Campaign from Vendor
                vendor_solution (dict): solution from vendor

            returns:
                result (bool): Should update campaign
        """

        # get solution_version_arn from vendor campaign
        vendor_campaign_solution_version_arn = self.get_vendor_campaign_solution_version_arn(
            vendor_campaign=vendor_campaign
        )

        # get solution version arn in database
        database_campaign_solution_version_arn = campaign['solution_version_arn']

        # get solution arn from vendor solution
        vendor_solution_version_arn = self.get_vendor_solution_version_arn(
            vendor_solution=vendor_solution
        )

        # get campaign status from vendor campaign
        vendor_campaign_status = self.get_vendor_campaign_status(
            vendor_campaign=vendor_campaign
        )

        # get solution status from vendor campaign
        latest_vendor_solution_version_status = self.get_vendor_solution_status(
            vendor_solution=vendor_solution
        )

        # don't update if mismatch in vendor and database solution version arns
        if (
                database_campaign_solution_version_arn !=
                vendor_campaign_solution_version_arn
        ):
            return False

        # Don't update campaign if its on the latest solution version
        if (
                vendor_solution_version_arn ==
                vendor_campaign_solution_version_arn
        ):
            return False

        if vendor_campaign_status in self.campaign_statuses['blocking']:
            return False

        # If the current campaign state == failed, update immediately
        if vendor_campaign_status in self.campaign_statuses['danger']:
            return True

        # Don't update campaign if the vendor solution version is not active
        if latest_vendor_solution_version_status not in self.solution_statuses['success']:
            return False

        return True

    def update_state(
            self,
            campaign: dict,
            solution: dict,
            **kwargs
    ):
        """
            Run the update workflow for campaigns

            Args:
                campaign (dict): database campaign config
                solution (dict): database solution config
            Return:
                (dict): new campaign config

        """
        client_code = kwargs.get("client_code", None)
        service_slug = kwargs.get("service_slug", None)
        engine_slug = kwargs.get("engine_slug", None)
        new_campaign = {
            'arn': None,
            'solution_version_arn': None,
            'status': None,
            'last_updated_at': None
        }

        # run vendor campaign update workflow
        response = self.sagemaker_update_campaign(
            endpoint_name=campaign.get('arn'),
            training_job_name=solution.get('arn'),
            container_image=campaign.get('container_image', None),
            client_code=client_code,
            service_slug=service_slug,
            engine_slug=engine_slug
        )

        new_campaign['arn'] = response.get('EndpointName')
        new_campaign['solution_version_arn'] = response.get('TrainingJobName')

        # get the current time as last updated at time
        current_time_at_update = datetime.strftime(
            datetime.now(),
            self.date_utils.get_mysql_date_format()
        )
        new_campaign.update({
            'last_updated_at': current_time_at_update
        })

        # set new campaign status as updating status
        new_campaign.update({
            'status': self.campaign_statuses['updating'][0]
        })

        # get next update time, this is not a required trigger here
        # deprecate this going forward
        next_update_time = new_campaign.get('next_update_time', None)
        if next_update_time is not None:
            # next update time not implemented
            new_campaign.update({
                'next_update_time': None
            })

        return new_campaign



"""
    Personalize recommendations manager module
"""
import json
from sdc_helpers.models.client import Client
import boto3
from sdc_engine_helpers.personalize.manager import Manager


class RecommendationsManager(Manager):
    """
        Manages retrieval of Personalize recommendations for a given set of parameters
    """
    # pylint: disable=raise-missing-from
    def get_recommendations(self):
        """
            Get the recommendations for the specified parameters

            returns:
                results (dict): Results from the Personalize invocation
        """
        client = self.get_client()

        item_id = self.parameters.get('itemId')

        if not item_id:
            raise Exception('ClientError: `itemId` is required')

        cache_key = 'recommend-{client_id}-{item_id}'.format(
            client_id=client.id,
            item_id=item_id
        )

        results = self.redis_helper.redis_get(key=cache_key)

        if not results:
            campaign_kwargs = {
                'recipe': self.parameters.get('recipe', 'related_items'),
                'event_type': self.parameters.get('eventType', 'all')
            }
            campaign, _ = self.query_helper.get_campaign(
                client_id=client.id,
                service_id=self.service.id,
                engine_slug=self.engine_slug,
                **campaign_kwargs
            )

            if not campaign:
                raise Exception('ServerError: Could not determine campaign for this client')

            campaign_arn = campaign['arn']
            default_results = campaign.get('default_results', [])

            results = self.get_results(
                client=client,
                campaign_arn=campaign_arn,
                default_results=default_results
            )

            self.redis_helper.redis_set(
                key=cache_key,
                value=json.dumps(results),
                expiry=7200
            )
        else:
            results = json.loads(results)

        return results

    def get_results(self, *, client: Client, campaign_arn: str, default_results: list) -> list:
        """
            Determine the results for this request

            args:
                client (Client): The client requesting the recommendation
                campaign_arn (str): The AWS Personalize campaign ARN
                default_results (list): The campaign's default results. If the recommendations
                                        are similar to the default results, an empty results
                                        list is returned as the default results are returned
                                        for an an item that does not exist in the solution the
                                        campaign is referencing

            returns:
                results (list): Results from the Personalize invocation

        """
        item_id = self.parameters.get('itemId')
        num_results = self.parameters.get('numResults', 25)

        results = []

        try:
            personalize_runtime = boto3.client('personalize-runtime')

            response = personalize_runtime.get_recommendations(
                campaignArn=campaign_arn,
                itemId=item_id,
                numResults=num_results
            )
        except Exception as ex:
            raise Exception('ServerError: {exception}'.format(exception=str(ex)))

        for item in response['itemList']:
            results.append(item['itemId'])

        # Check if the results are the same or similar to the default results
        # and return empty array if so
        intersection = set(default_results).intersection(results)

        if (
                len(results) == len(intersection) or
                len(intersection) / len(results) >= 0.8
        ):
            if len(results) != len(intersection):
                print(
                    'Results for {client_name} are similar but not the same to the default results '
                    'which probably means default results are out of sync. Saving default results'
                    .format(client_name=client.name)
                )

            results = []

        return results



"""
    Dataset state manager module
        - state managers have knowledge of state transition and meaning
        - state manager is called to check state
        - state manager is called to transition state
        - state manager only sends back True/False. Do something or don't.

        may want to:
        - add validation of complex types: Url, etc https://validators.readthedocs.io/en/latest/
"""
import io
import json
from datetime import datetime
import boto3
from botocore.exceptions import ClientError as botocore_clienterror
import sdc_helpers.utils as utils
from sdc_helpers.redis_helper import RedisHelper
from sdc_engine_helpers.sagemaker.utils import parse_s3_url
from sdc_engine_helpers.date_utils import DateUtils

# pylint: disable=too-few-public-methods
class DatasetStateManager:
    """
        Given a database dataset object,

        1. get_state
            - check whether dataset should be updated by
                call vendor API's to check if campaign solution
                version differs from dataset solution version
            OR
            - check if the dataset status in the database needs to be refreshed.
        2. update_state
            - call vendor API's to execute update workflow for dataset
            - update status of dataset in database

    """
    date_utils = DateUtils()
    redis_helper = None
    query_helper = None
    # vendor
    vendor_campaign_arn_path = 'EndpointName'
    vendor_campaign_status_path = 'EndpointStatus'
    # input data key in TrainingJob
    vendor_campaign_solution_arn_path = 'ProductionVariants.0.VariantName'

    campaign_statuses = {
        'running':['Creating', 'Updating', 'SystemUpdating', 'Deleting'],
        'failed':['Failed', 'RollingBack'],
        'danger':['Failed'],
        'blocking':[
            'Creating',
            'Updating',
            'SystemUpdating',
            'Deleting',
            'Failed',
            'RollingBack'],
        'stopped':['OutOfService'],
        'active':['InService', None, 'None']
    }

    solution_statuses = {
        'blocking':['InProgress', 'Stopping', 'Failed', 'Stopped'],
        'success':['Completed', None, 'None']
    }

    dataset_statuses = {
        'active': ['Active', None, 'None'],
        'blocking': ['Updating']
    }

    def __init__(self, **kwargs):
        self.sagemaker = boto3.client('sagemaker')
        self.s3_client = boto3.client('s3')
        redis_config = kwargs.get('redis_config', {})
        if not isinstance(redis_config, dict):
            raise TypeError('Only dict is supported for redis_config')
        self.redis_helper = RedisHelper(**redis_config)

    @staticmethod
    def get_vendor_solution_dataset_identifier(
            *,
            vendor_solution: dict,
            dataset_label: str,
            file_type: str
        ) -> str:
        """
            Get a Sagemaker Training job input dataset by channel name

            args:
                vendor_solution (dict): Solution from Personalize
                dataset_label (str): dataset label as stored in db

            returns:
                status (str): Sagemaker solution version status

        """
        input_configs = vendor_solution.get("InputDataConfig")
        if input_configs is not None:
            for d_configs in input_configs:
                channel_name = d_configs.get("ChannelName")

                if channel_name is None:
                    raise KeyError("ChannelName must be provided by vendor")

                if channel_name == dataset_label:
                    s3_folder_path = utils.dict_query(
                        dictionary=d_configs,
                        path='DataSource.S3DataSource.S3Uri'
                    )
                    return "{}/{}.{}".format(
                        s3_folder_path,
                        dataset_label,
                        file_type
                    )

        else:
            raise ValueError("InputDataConfig is missing, vendor should provide this.")

        raise KeyError((
            "Channel with name = {} was not"
            "found in vendor solution".format(dataset_label)
        ))

    def download_s3_blob(self, s3_url: str):
        """download blob from s3"""
        print("Starting.. Download blob")
        # parse s3 url -> bucket, file_path
        bucket, file_path = parse_s3_url(s3_url)
        # load data from s3 as byte steam
        data_stream = io.BytesIO()
        try:
            self.s3_client.download_fileobj(
                bucket,
                file_path,
                data_stream
            )
            data_stream.seek(0)

            return json.loads(
                data_stream.read().decode("UTF-8")
            )

        except botocore_clienterror as exception:
            if exception.response['Error']['Code'] == "404":
                # The object does not exist.
                print((
                    "Unable to find s3 object. Skipping the task, "
                    "the following exception was found = {}".format(exception)
                ))
                return None
            raise

    def upload_s3_blob(self, blob: io.BytesIO, s3_url: str):
        """download blob from s3"""
        # parse s3 url -> bucket, file_path
        bucket, file_path = parse_s3_url(s3_url)
        try:
            # upload to s3
            self.s3_client.upload_fileobj(
                blob, bucket, file_path
            )
        except botocore_clienterror as exception:
            if exception.response['Error']['Code'] == "404":
                # The object does not exist.
                print((
                    "Unable to find s3 object. Skipping the task, "
                    "the following exception was found = {}".format(exception)
                ))
            else:
                raise

    def execute_s3_versioning(self, json_data: dict, dst_url: str):
        """
            Streams data from source s3 url to a destination s3 url

            Args:
                src_url (str): source s3 url
                dst_url (str): destination s3 url
        """
        data_stream = io.BytesIO(
            json.dumps(json_data).encode("UTF-8")
        )

        self.upload_s3_blob(
            blob=data_stream,
            s3_url=dst_url
        )

    def delete_cached_dataset(self, cache_key: str):
        """deletes a cached dataset given a key

        Args:
            cache_key (str): dataset key
        """
        if self.redis_helper.redis_hexists(hashkey=cache_key) == 1:
            print("Deleting cached dataset")
            self.redis_helper.redis_delete(keys=[cache_key])

    def cache_lookup(self, json_data: dict, cache_key: str):
        """cache lookup in elastic cache

        Args:
            cache_key (str): key to use when caching
            json_data (dict): data to cache which should be simple dictionary lookup
        """
        if not isinstance(json_data, dict):
            raise Exception("ClientError: Json data is not a dictionary")

        if self.redis_helper.redis_hexists(hashkey=cache_key) == 1:
            print("This cache already exists. Deleting before caching new data")
            self.redis_helper.redis_delete(keys=[cache_key])

        self.redis_helper.redis_hset(
            hashkey=cache_key,
            dict_obj=json_data
        )

    def get_vendor_campaign_solution_version_arn(self, vendor_campaign: dict):
        """
            Get vendor campaign solution version status

            args:
                vendor_campaign (dict): Solution from vendor campaign

            returns:
                status (str): vendor solution version status

        """

        return utils.dict_query(
            dictionary=vendor_campaign,
            path=self.vendor_campaign_solution_arn_path
        )

    def get_vendor_campaign_status(self, vendor_campaign: dict):
        """
            Get vendor campaign solution version status

            args:
                vendor_campaign (dict): Solution from vendor campaign

            returns:
                status (str): vendor solution version status

        """

        return utils.dict_query(
            dictionary=vendor_campaign,
            path=self.vendor_campaign_status_path
        )

    def get_state(
            self, *,
            dataset: dict,
            campaign: dict):
        """
            Get dataset state

            Args:
                dataset (dict): database dataset config
                campaign (dict): database campaign config
            Return:
                (dict): current state config
        """
        state = {
            'should_update': False
        }

        # get current campaign from vendor
        vendor_campaign = self.sagemaker.describe_endpoint(
            EndpointName=campaign.get('arn')
        )

        # check if should update
        state['should_update'] = self.should_update(
            dataset=dataset,
            vendor_campaign=vendor_campaign
        )

        if state['should_update']:
            # set a new version status as InProgress
            state.update(
                {
                    'new_version_status': 'Updating'
                }
            )

        return state

    def update_state(
            self, *,
            dataset: dict,
            campaign: dict,
            cache_key: str
    ):
        """
            This is where jobs to create datasets are run

            Args:
                dataset (dict): database dataset config
                campaign (dict): database campaign config
            Return:
                (dict): new dataset
        """

        new_dataset = {
            'data': None,
            'identifier': None,
            'solution_version_arn': None,
            'last_updated_at': None,
            'last_status': None
        }

        # get vendor solution
        vendor_campaign = self.sagemaker.describe_endpoint(
            EndpointName=campaign.get('arn')
        )

        # get vendor campaign
        latest_vendor_solution_version_arn = self.get_vendor_campaign_solution_version_arn(
            vendor_campaign
        )

        vendor_solution = self.sagemaker.describe_training_job(
            TrainingJobName=latest_vendor_solution_version_arn
        )

        # update solution arn
        new_dataset.update({
            'solution_version_arn': latest_vendor_solution_version_arn
        })

        # update identifier
        new_dataset.update({
            'identifier': self.get_vendor_solution_dataset_identifier(
                vendor_solution=vendor_solution,
                dataset_label=dataset['label'],
                file_type=dataset['file_type']
            )
        })

        if dataset.get('should_store'):
            # update data
            identifier = new_dataset.get('identifier', None)
            # get dataset from reference
            if isinstance(identifier, str):
                print("Build prod destination url")
                prod_url = identifier.replace("/train/", "/prod/")
                print("Build previous destination url")
                previous_url = identifier.replace("/train/", "/previous/")

                print("Download dataset from prod/ in s3")

                json_data = self.download_s3_blob(s3_url=prod_url)
                if isinstance(json_data, dict):
                    print("Pushing to previous/ to s3")
                    self.execute_s3_versioning(
                        json_data=json_data,
                        dst_url=previous_url
                    )

            print("Download dataset from train/ from s3")
            # get data
            json_data = self.download_s3_blob(s3_url=identifier)

            if isinstance(json_data, dict):
                tmp_cache_clone = "{}-tmp-clone".format(cache_key)
                print("Creating a dataset clone")
                self.cache_lookup(
                    json_data=json_data,
                    cache_key=tmp_cache_clone
                )
                # cache to elastic cache
                print("Caching new prod dataset")
                self.cache_lookup(
                    json_data=json_data,
                    cache_key=cache_key
                )
                print("Removing clone")
                self.delete_cached_dataset(
                    cache_key=tmp_cache_clone
                )
                print("Pushing dataset to prod/ in s3")
                self.execute_s3_versioning(
                    json_data=json_data,
                    dst_url=prod_url
                )

        # date
        new_dataset.update({
            'last_updated_at': datetime.strftime(
                datetime.now(),
                self.date_utils.get_mysql_date_format()
                )
        })

        # update status
        new_dataset.update({
            'last_status': 'Active'
        })

        return new_dataset

    def should_update(
            self,
            *,
            dataset: dict,
            vendor_campaign: dict
    ) -> bool:
        """Check if dataset should update"""
        # (look at campaign state manager)
        latest_vendor_campaign_solution_version_arn = self.get_vendor_campaign_solution_version_arn(
            vendor_campaign
        )

        latest_vendor_campaign_status = self.get_vendor_campaign_status(
            vendor_campaign
        )

        # Don't update dataset if its on the latest solution version
        if (
                dataset.get('solution_version_arn') ==
                latest_vendor_campaign_solution_version_arn
        ):
            return False

        if latest_vendor_campaign_status not in self.campaign_statuses['active']:
            return False

        if dataset.get('last_status') not in self.dataset_statuses['active']:
            return False

        return True



"""
    Solution state manager module
"""
from datetime import datetime
import sdc_helpers.utils as utils
import boto3
from sdc_engine_helpers.date_utils import DateUtils


class SolutionStateManager:
    """
        Given a database solution object, check whether a new solution version needs to be
        created OR whether the status in the database needs to be refreshed
    """
    personalize = None
    date_utils = None

    def __init__(self):
        self.personalize = boto3.client('personalize')
        self.date_utils = DateUtils()

    def get_state(self, *, solution: dict) -> dict:
        """
            Get the current state of the solution in Personalize

            args:
                solution (dict): Solution from the database

            returns:
                state (dict): Dictionary containing:
                              1) should_create_version
                              2) should_refresh_version_status
                              3) next_version_creation_time (if frequency specified)
                              4) new_version_status

        """
        state = {
            'should_create_version': False,
            'should_refresh_version_status': False,
            'next_version_creation_time': None,
            'new_version_status': None
        }

        # Get Personalize solution from the Personalize client, an exception is thrown if
        # the solution arn is not found
        personalize_solution_response = self.personalize.describe_solution(
            solutionArn=solution.get('arn')
        )
        personalize_solution = personalize_solution_response.get('solution')

        # Check for creation
        state['should_create_version'] = self.should_create_solution_version(
            solution=solution,
            personalize_solution=personalize_solution
        )

        if state['should_create_version']:
            this_version_creation_time = solution.get('next_version_creation_time', None)
            version_creation_frequency = solution.get('version_creation_frequency', None)

            if version_creation_frequency:
                state['next_version_creation_time'] = self.date_utils.get_next_time(
                    this_time=this_version_creation_time,
                    frequency=version_creation_frequency
                )
        else:
            # Check if the database version status needs to be updated
            version_status = solution.get('version_status', None)
            personalize_solution_status = self.get_personalize_version_status(
                personalize_solution=personalize_solution
            )

            if personalize_solution_status != version_status:
                state['should_refresh_version_status'] = True
                state['new_version_status'] = personalize_solution_status

        return state

    @staticmethod
    def should_create_solution_version(*, solution: dict, personalize_solution: dict) -> bool:
        """
            Check whether a solution version should be created when all of the following are true:

            1) The latest solution version in Personalize is ACTIVE or FAILED
            2) The current time is >= the scheduled next version creation time (if specified)

            args:
                solution (dict): Solution from the database
                personalize_solution (dict): Solution from Personalize

            returns:
                result (bool): Should create solution version

        """
        # If the latest solution version is still creating in Personalize, don't create
        personalize_solution_version_status = utils.dict_query(
            dictionary=personalize_solution,
            path='latestSolutionVersion.status'
        )

        if personalize_solution_version_status not in ['ACTIVE', 'CREATE FAILED']:
            return False

        # If the Personalize solution version creation failed, create a new version immediately
        if personalize_solution_version_status == 'CREATE FAILED':
            return True

        version_creation_frequency = solution.get('version_creation_frequency', None)

        if version_creation_frequency:
            next_version_creation_time = solution.get('next_version_creation_time')
            if next_version_creation_time:
                if (
                        datetime.now() <
                        datetime.strptime(next_version_creation_time, "%Y-%m-%d %H:%M:%S")
                ):
                    return False

        return True

    @staticmethod
    def get_personalize_version_status(*, personalize_solution: dict) -> str:
        """
            Get Personalize solution version status

            args:
                personalize_solution (dict): Solution from Personalize

            returns:
                status (str): Personalize solution version status

        """

        return utils.dict_query(
            dictionary=personalize_solution,
            path='latestSolutionVersion.status'
        )



"""
    Sagemaker recommendations manager module
"""
import json
from sdc_helpers.models.client import Client
from sdc_helpers.redis_helper import RedisHelper
import boto3
from sdc_engine_helpers.sagemaker.utils import create_resource_name
from sdc_engine_helpers.sagemaker.sagemaker_query_helpers import \
    SagemakerEngineHelpers

class RecommendationsManager:
    """
        Manages retrieval of Sagemaker recommendations for a given set of parameters
    """
    # pylint: disable=raise-missing-from
    client = {}
    engine = None

    def __init__(self, **kwargs):
        """
            kwargs:
                service_slug (str): slug of service to use
                rds_config (dict): config for mysql database
                redis_config (dict): config for redis cache database
                item_ids (str): name of item id lookup in mysql db
                item_codes (str): name of item code lookup in mysql db
        """
        # get rds config
        rds_config = kwargs.get('rds_config', {})
        if not isinstance(rds_config, dict):
            raise TypeError('Only dict is supported for rds_config')

        # get redis config
        redis_config = kwargs.get('redis_config', {})
        if not isinstance(redis_config, dict):
            raise TypeError('Only dict is supported for redis_config')

        # instantiate query helpers
        self.query_helper = SagemakerEngineHelpers(
            rds_config=rds_config,
            redis_config=redis_config
        )

        self.redis_helper = RedisHelper(**redis_config)
        # service slug is the same for all engines
        service_slug = kwargs.pop('service_slug', 'recommend')

        # get name of lookups
        self.item_id_identifier = kwargs.pop('item_id_identifier', 'item_id_lookup')
        self.item_code_identifier = kwargs.pop('item_code_identifier', 'item_code_lookup')

        # load service details from service table
        self.service = self.query_helper.get_service(slug=service_slug)

    def __del__(self):
        """
            Close connections on delete
        """
        del self.query_helper
        del self.redis_helper

    def get_recommendations(
            self,
            *,
            client_id: id,
            item_id: str,
            engine: str,
            **kwargs
    ):
        """
            Get the recommendations for the specified parameters

            parameters:
                - dictionary of parameter arguments
            returns:
                results (dict): Results from the Sagemaker invocation
        """
        results = None
        self.client = self.get_client(client_id=client_id)
        self.engine = engine
        # recipe is set by client config
        recipe = kwargs.get('recipe', 'related_items')
        num_results = kwargs.get('num_results', 25)
        from_cache = kwargs.get('from_cache', True)

        cache_key = (
            'recommendation-request-'\
            '{item_id}-{service_id}-'\
            '{client_id}-{engine_slug}'.format(
                item_id=item_id,
                service_id=self.service.id,
                client_id=self.client.id,
                engine_slug=self.engine
            )
        )

        if from_cache:
            results = self.redis_helper.redis_get(key=cache_key)

        if not results:
            campaign = self.get_campaign(
                client_id=self.client.id,
                engine=self.engine,
                recipe=recipe
            )

            if not campaign:
                raise Exception('ServerError: Could not determine campaign for this client')

            # changed campaign_arn to be campaign_name
            endpoint_name = campaign['arn']

            results = self.get_results(
                item_id=item_id,
                endpoint_name=endpoint_name,
                num_results=num_results
            )
            self.redis_helper.redis_set(
                key=cache_key,
                value=json.dumps(results),
                expiry=7200
            )
        else:
            results = json.loads(results)

        return results

    def get_client(
            self, *,
            client_id: int = None,
            api_key_id: str = None
    ) -> Client:
        """
            Determine the client with the supplied parameters

            Expects either client_id or api_key_id
            args:
                client_id (int): client id to fetch
                api_key_id (int): api_key_id to fetch client_id

            returns:
                client (Client): The determined client

        """
        client = None

        if client_id is not None:
            client = self.query_helper.get_client(client_id=client_id)
        else:
            if api_key_id is not None:
                client = self.query_helper.get_client(api_key_id=api_key_id)

        if client is None:
            raise Exception('ClientError: Could not determine client for this request')

        return client

    def get_campaign(
            self, *,
            client_id: int,
            engine: str,
            recipe='related_items'
    ) -> dict:
        """
            Determine the campaign with the supplied parameters
            note:
                - Currently: Campaign arn = (Sagemaker Engpoint Name)

            args:
                client_id (int): client id
                engine (str): engine slug
                recipe (str): recipe name

            returns:
                campaign (dict): Campaign from the database

        """
        campaign_kwargs = {'recipe': recipe}

        campaign, _ = self.query_helper.get_campaign(
            client_id=client_id,
            service_id=self.service.id,
            engine_slug=engine,
            from_cache=True,
            **campaign_kwargs
        )

        return campaign

    def encode_item_code(
            self,
            *,
            item_ids: list
    ):
        """
            Lookup a list of item codes

            args:
                 item_ids (list) : list of item id's
                 from_cache (bool) : Get lookup from cache first
            return:
                item_codes (list) : list of item codes
        """
        item_codes = []

        cache_key = create_resource_name(
            context='recommendation',
            function=self.item_code_identifier,
            additional_details=[
                self.service.id, self.client.id, self.engine
            ]
        )

        # check that key exists
        key_exists = self.redis_helper.redis_hexists(
            hashkey=cache_key
        )

        if key_exists == 0:
            tmp_cache_clone = "{}-tmp-clone".format(cache_key)
            # check that key exists
            key_exists = self.redis_helper.redis_hexists(
                hashkey=tmp_cache_clone
            )
            if key_exists == 0:
                raise Exception((
                    "ServerError: Item Code lookup could "
                    "not be found in cache. Both main and clone missing."
                ))
            # use clone as lookup
            cache_key = tmp_cache_clone

        # encode item_id -> item_code
        item_codes = self.redis_helper.redis_hmget(
            hashkey=cache_key,
            keys=item_ids
        )

        results = []
        for item_code in item_codes:
            # decode item_ids bytes -> str
            if isinstance(item_code, bytes):
                item_code = item_code.decode("utf-8")

            # check if item_id is none
            if item_code is not None:
                results.append(item_code)

        return results

    def decode_item_code(
            self,
            *,
            item_codes: list
    ):
        """"
            Lookup a list of item ids

            args:
                 item_codes (list) : list of item codes
                 from_cache (bool) : Get lookup from cache first
            return:
                item_ids (list) : list of item id
        """
        item_ids = []

        cache_key = create_resource_name(
            context='recommendation',
            function=self.item_id_identifier,
            additional_details=[self.service.id, self.client.id, self.engine]
        )
        # check that key exists
        key_exists = self.redis_helper.redis_hexists(
            hashkey=cache_key
        )

        if key_exists == 0:
            tmp_cache_clone = "{}-tmp-clone".format(cache_key)

            # check that key exists
            key_exists = self.redis_helper.redis_hexists(
                hashkey=tmp_cache_clone
            )

            if key_exists == 0:
                raise Exception((
                    "ServerError: Item ID lookup could "
                    "not be found in cache. Both main and clone missing."
                ))
            # use clone as lookup
            cache_key = tmp_cache_clone

        # if from_cache:
        # decode item_codes -> item_ids
        item_ids = self.redis_helper.redis_hmget(
            hashkey=cache_key,
            keys=item_codes
        )

        results = []
        for item_id in item_ids:
            if isinstance(item_id, bytes):
                # decode item_ids bytes -> str
                item_id = item_id.decode("utf-8")

            # check if item_id is none
            if item_id is not None:
                results.append(item_id)

        return results

    def get_results(
            self,
            *,
            item_id: str,
            endpoint_name: str,
            num_results: int = 25
    ) -> list:
        """
            Determine the results for this request

            args:
                item_id (str): The client requesting the recommendation
                endpoint_name (str): The AWS Sagemaker endpoint_name
                num_results (int) : Number of recommendation results to return

            returns:
                results (list): Sagemaker invocation results

        """
        results = []
        # lookup item code
        item_codes = self.encode_item_code(item_ids=[item_id])
        # test that item_code is valid
        if len(item_codes) == 0:
            raise Exception('ClientError: item_id was not found in lookup')

        try:
            request_item = int(item_codes[0])
            sagemaker_runtime = boto3.client('sagemaker-runtime')
            # make the recommender ignore self recommendation
            event = {
                'item_code': request_item,
                'args': {
                    'n_similar': num_results,
                    'exclude': [
                        request_item
                    ]
                }
            }
            payload = json.dumps(event)

            # get recommendations
            response = sagemaker_runtime.invoke_endpoint(
                EndpointName=endpoint_name,
                ContentType='application/json',
                Body=payload
            )

            # decode application/json response
            response_decoded = json.loads(response['Body'].read().decode())


            if isinstance(response_decoded, str):
                # if decorded response still string
                # assume error and raise
                raise Exception(response_decoded)

            #extract itemcodes
            recommendations = [r['item_code'] for r in response_decoded]

        except sagemaker_runtime.exceptions.ModelError:
            # insure model errors are returned as 500
            msg = "ServerError: 500 Internal Server Error"
            raise Exception(msg)

        # decode recommendations bytes -> str
        results = self.decode_item_code(item_codes=recommendations)
        # check list of item_ids are not all None
        if len(results) == 0:
            raise Exception('ClientError: Unable to decode recommendations to item_id')

        return results
